#include <stdio.h>
#include <stdarg.h>
#include <unistd.h>
#include <signal.h>
#include <setjmp.h>

#include "tetris.h"
#include "video.h"
#include "cons.h"

#include "main.h"
#include "music.h"

#include "music_cv1d.h"
#include "music_cv2b.h"
#include "music_tet1.h"
#include "music_tet4.h"
#include "music_tetb.h"
#include "music_tetk.h"
#include "music_tetl.h"
#include "music_tett.h"
#include "music_teta.h"
#include "music_tetr.h"
#include "music_teto.h"

static jmp_buf ResetGame;
static tetris_t Game, Game2;

void SetText(int x,int y, int c,int tr,int time, const char *fmt, ...)
{
	char Buf[512], *s;
	va_list ap;
	int nx=x;
	
	va_start(ap, fmt);	
	vsprintf(Buf, fmt, ap);
	va_end(ap);
	
	for(s=Buf; *s; s++)
		switch(*s)
		{
			case '\b': x--; break;
			case '\n': y++;
			case '\r': x=nx; break;
			default:
				SetChar(x,y,*s, c,tr,time);
				x++;
		}
}

void Center(int y, int c,int tr,int time, const char *s)
{
	SetText(20-strlen(s)/2,y, c,tr,time, "%s", s);
}

static void DrawMainBorder(void)
{
	int x, y;
	for(y=0; y<24; y++)
	{
		SetChar( 0,y, corner(c_up+c_dn+c_rt),mainborder, 0,0);
		SetChar( 1,y, corner(c_up+c_dn+c_lt),mainborder, 0,0);
		SetChar(38,y, corner(c_up+c_dn+c_rt),mainborder, 0,0);
		SetChar(39,y, corner(c_up+c_dn+c_lt),mainborder, 0,0);
	}
	for(x=1; x<39; x++)
	{
		SetChar(x, 0, corner(c_lt+c_rt),mainborder, 0,0);
		SetChar(x,23, corner(c_lt+c_rt),mainborder, 0,0);
	}
}

static void DrawScoreBorder(void)
{
	int x, y;
	
	for(x=20-9; x<20+9; x++)
		SetChar(x,4, corner(c_lt+c_rt), scoreborder,0,0);
			
	for(y=1; y<=4; y++)
	{
		SetChar(20-9,y, corner(c_up+(y<4?c_dn:c_rt)), scoreborder,0,0);
		SetChar(20+9,y, corner(c_up+(y<4?c_dn:c_lt)), scoreborder,0,0);
		SetChar(20-3,y, corner(c_up+(y<4?c_dn:c_lt+c_rt)), scoreborder,0,0);
		SetChar(20+3,y, corner(c_up+(y<4?c_dn:c_lt+c_rt)), scoreborder,0,0);
	}
	
	SetText(20-2,1, scorecolor,0,0, "Score\nLines\nLevel");
}

static void GameBorder(void)
{
	int y, x;
	
	Center(20, 7,2,0, "                              ");
	Center(21, 7,2,0, "                           ");
	
	for(x=0; x<rakepax; x++)
	{
		SetChar(12+x,23, corner(c_lt+c_rt+c_up), mainborder,2,0);
		SetChar(27-x,23, corner(c_lt+c_rt+c_up), mainborder,2,0);
	}
	for(y=5+x; y<23; y++)
		for(x=0; x<rakepax; x++)
		{
			SetChar(12+x,y, corner(c_up+c_dn), rakennelma,2,0);
			SetChar(27-x,y, corner(c_up+c_dn), rakennelma,2,0);
		}
	for(y=0; y<rakepax; y++)
		for(x=12; x<=27; x++)
		{
			int c = (x<12+y||x>27-y)
				?	c_up+c_dn
				:	(x>12+y?c_rt:c_dn) + (x<27-y?c_lt:c_dn);
			SetChar(x,5+y, corner(c), rakennelma,2,0);
		}
}

static int Tick(int numticks)
{
	UpdateVideo(numticks);
	TickMusa(numticks);
	return numticks;
}

static int WhichKey(void)
{
	int c = getch();
	if(c=='A'||c=='w'||c=='H')c='';
	if(c=='B'||c=='s'||c=='P')c='';
	if(c=='D'||c=='a'||c=='K')c='';
	if(c=='C'||c=='d'||c=='M')c='';
	if(c=='\r')c='\n';
	return c;
}

static void MusicReactor(int MusicNum)
{
	switch(MusicNum)
	{
		case -2:StartMusa(tet1); break; /* main */
		case -1:break;					/* enter key */
		case 0:	StartSilence(); break;
		case 1: StartMusa(tet4); break;
		case 2: StartMusa(cv1d); break;
		case 3: StartMusa(cv2b); break;
		case 4: StartMusa(tetl); break;
		case 5: StartMusa(tetb); break;
		case 6: StartMusa(tetk); break;
		case 7: StartMusa(tett); break;
		case 8: StartMusa(teta); break;
		case 9: StartMusa(tetr); break;
	}
}

BEGIN_EFFECT(eff_move, CHANNEL 0, TEMPO 400, LENGTH 4)
	50, 50,
	54, 50,
	60, 50,
	74, 50
END_EFFECT;

BEGIN_EFFECT(eff_sel, CHANNEL 0, TEMPO 300, LENGTH 6)
	50, 50,
	47, 45,
	44, 40,
	41, 35,
	38, 30,
	35, 25
END_EFFECT;

/* Reactor args:
 *      -1=affirmated
 *   other=selection no
 */
static int Menu(void (*Reactor)(int), const char *Title, ...)
{
	int a, SelCount, Cur, Longest, x, Ticks, top=11;
	va_list ap;
	
	va_start(ap, Title);
	for(Longest=0, x=0;;)
	{
		char *s = va_arg(ap, char *);
		int Len;
		
		if(x++ > 8)top--;
		
		if(!s)break;
		Len=strlen(s);
		if(Len>Longest)Longest=Len;
	}
	va_end(ap);
	
	/* Menu is transparent - good? */
	Center(top-2, menucolor,1,0, Title);
	
	x = 21 - Longest/2;
	
	va_start(ap, Title);
	for(SelCount=0;;)
	{
		char *s = va_arg(ap, char *);
		if(!s)break;
		
		SetText(x, top+SelCount++, menucolor,1,0, "%s", s);
	}
	va_end(ap);
	
	Cur=0;
	
Redraw:
	if(Reactor)Reactor(Cur);
	for(a=0; a<SelCount; a++)
	{
		SetChar(x-3, top+a, a==Cur?'-':' ', menuarrow,1,0);
		SetChar(x-2, top+a, a==Cur?'>':' ', menuarrow,1,0);
	}
	for(;;)
	{
		Ticks=0;
		while(!kbhit())
		{
			Tick(1);
			if(++Ticks >= 10 SECONDS)
				longjmp(ResetGame, 1);
		}
		switch(WhichKey())
		{
			case '':	
				Cur=(Cur+SelCount-1)%SelCount;
				PlayEff(eff_move);
				goto Redraw;
			case '':
				Cur=(Cur+         1)%SelCount;
				PlayEff(eff_move);
				goto Redraw;
			case '':
			case '':
				PlayEff(eff_move);
				break;
			case '\n':
				PlayEff(eff_sel);
				if(Reactor)Reactor(-1);
				for(a=-2; a<SelCount; a++)
					SetText(2,top+a, 7,1,0, "%36s", "");
				return Cur;
		}
	}
}

void CtrlC(int dummy)
{
	DoneVideo();
	exit(dummy);
}

int main(int argc, char **argv)
{
	int GameMode;
	int Handicap;
	int Level;
	int Music;
	int KeyMode;
	
	#define KEYSETUPCOUNT 5
	
	static const char *KeySetups[KEYSETUPCOUNT] =
	{
		"rotate1,drop,stepdown",
		"rotate1,rotate2,stepdown",
		"rotate2,rotate1,drop",
		"rotate2,rotate1,stepdown",
		"rotate2,drop,stepdown"
	};
	static const int Key[KEYSETUPCOUNT][3] =
	{	
		{KEY_TURN1, KEY_DROP, KEY_DOWN},
		{KEY_TURN1, KEY_TURN2, KEY_DOWN},
		{KEY_TURN2, KEY_TURN1, KEY_DROP},
		{KEY_TURN2, KEY_TURN1, KEY_DOWN},
		{KEY_TURN2, KEY_DROP, KEY_DOWN}
	};	
	
	argc=argc;
	argv=argv;
	
	signal(SIGINT, CtrlC);
	
	Randomize();
	
	InitVideo();
	
	if(setjmp(ResetGame))
		SetText(0,0, 8,1,2 SECONDS, "reset");
	
	ClrScr();
	
	MusicReactor(-2);
	
	Center(8, 12,1,0, "JOELTRIS 10");
	Center(9,  4,1,0.5 SECONDS, VERSION);
	Center(14,12,1,0, "PRESS ENTER");
	Center(20, 1,1,2.5 SECONDS, "Written by Bisqwit");
	Center(21, 1,1,2.5 SECONDS, "http://iki.fi/bisqwit/");
	
	{int sx[starcount], sy[starcount], sd[starcount], sp[starcount], a;
	for(a=starcount; a--; sd[a]=0)
	{
		sx[a]=rand()%40;
		sy[a]=rand()%24;
		sp[a]=rand()&3;
		if(!sp[a])sp[a]=rand()&3;
	}	
	for(;;)
	{
		while(!kbhit())
		{
			for(a=0; a<starcount; a++)
				if(sd[a])sd[a]--;
				else
				{
					static const int C[] = {15,7,7,8};
					sd[a]=sp[a];
					if(QColor[sy[a]][sx[a]] != 1
					&& QField[sy[a]][sx[a]] == '-')
						SetChar(sx[a],sy[a], ' ',7,1,0);
						
					if(--sx[a] < 0)sx[a]=39, sy[a]=rand()%24;
						
					if(QColor[sy[a]][sx[a]] == 7
					&& QField[sy[a]][sx[a]] == ' ')
						SetChar(sx[a],sy[a], '-',C[sp[a]],1,0);
				}
			Tick(1);
		}			
		if(WhichKey()=='\n')break;
	}}
	#undef starcount
	
Over:
	ClrScr(); /* Wipe out the stars */
	DrawMainBorder();
	
	Center(8,  7,1,0, "           ");
	Center(14, 7,1,0, "           ");
	
	DrawScoreBorder();

	StartSilence();	/* Stop any music */
	
	GameMode = Menu(
		NULL,
		"GAME SELECT",
		"1 player",
		"versus computer",
		"demo",
		NULL);
		
	KeyMode = Menu(
		NULL,
		"KEYBOARD BINDINGS (UP, SPACE & DOWN)",
		KeySetups[0],
		KeySetups[1],
		KeySetups[2],
		KeySetups[3],
		KeySetups[4],
		NULL);
		
	Level = Menu(
		NULL,
		"LEVEL SELECT",
		"0", "1", "2", "3", "4", "5", "6", "7", "8", "9",
		NULL);
		
	Handicap = Menu(
		NULL,
		"HANDICAP SELECT",
		"0", "3", "6", "9", "12",
		NULL) * 3;
		
	Center(20, 1,1,3 SECONDS, "Some of the music melodies are");
	Center(21, 1,1,3 SECONDS, "(C) Nintendo and (C) Konami");

	Music = Menu(
		MusicReactor,
		"MUSIC SELECT",
		"Silence",
		"Dance",
		"Frog",
		"Mansion",
		"Loginska",
		"Bradinsky",
		"Karinka",
		"Troika",
		"ASO Tetris theme",
		"ASO Technotris",
		NULL);

	GameBorder();
	
	switch(GameMode)
	{
		case 0:	/* 1 player */
			InitTetris(&Game,  Level,Handicap,  2,5,  2,12, 0);
			break;
		case 1:	/* versus computer */
			InitTetris(&Game,  Level,Handicap,  2,5,  2,12, 0);
			InitTetris(&Game2, Level,Handicap, 28,5, 34,24, 1);
			break;
	}
	
	MusicReactor(Music);
	
	for(;;)
	{
		switch(GameMode)
		{
			case 0:
				TickTetris(&Game, 1);
				break;
			case 1:
				TickTetris(&Game, 1);
				TickTetris(&Game2, 1);
				break;
		}
		
		while(kbhit())
		{
			int a = WhichKey();
			
			switch(a)
			{
				case '':
					a = KEY_LT;
					break;
				case '':
					a = KEY_RT;
					break;
				case '':
					RefreshScreen(0);
					break;
				case '\n':
				case 'p':
					if(Game.State == S_OVER)goto Over;
					a = KEY_PAUSE;
					break;
				case '':
					a = Key[KeyMode][0];
					break;
				case ' ':
					a = Key[KeyMode][1];
					break;
				case '':
					a = Key[KeyMode][2];
					break;
				default:
					/* Do not confuse the game */
					a=0;
					break;
			}
			
			if(a)
				GrantKey(&Game, a);
		}
		
		Tick(1);
	}
	
	KillTetris(&Game);
	KillTetris(&Game2);
	
	DoneMusa();
	DoneVideo();
	
	return 0;
}
