/*
 * Video implementation for vt100 terminals
 *
 */

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <signal.h>

#include "video.h"
#include "cons.h"

#define EVERYFLUSH 0

/* Window upperleft coordinates on screen */
#define XA ((COLS -40)/2)
#define YA ((LINES-24)/2)
	
static int TextAttr;
static int OldAttr;
static int WhereX, WhereY;

/* I hope this is big enough! */
static char OutputBuffer[16384];
static char *OutputTail = OutputBuffer;

/* This is for speed */
static char ColorDiffLen[256][256];

static void Buffaa(const char *s)
{
	/* Todo: Check buffer overflow, call Flush() if necessary */
	
	while(*s)*OutputTail++ = *s++;
}
static void Flush(void)
{
	*OutputTail = 0;
	fputs(OutputBuffer, stdout);
	fflush(stdout);
	#if EVERYFLUSH
	usleep(10000);
	#endif
	OutputTail = OutputBuffer;
}

#define SetAttr(n) (TextAttr=(n))
#define Colors 1 /* Yes, we want colors */

#ifndef DJGPP
static char *ColorDiff(int old, int new)
{
	static char Buff[16];
	char *be = Buff;
	
	/* Todo:
	 * If the background color of old and new is the same,
	 * and is writing only spaces, there is no need to
	 * change the colour - it will still look the same.
	 */
	
    if(TextAttr == OldAttr || !Colors)
    {
    	Buff[0] = 0;
    	return Buff;
    }
	be[0] = 27;
	be[1] = '[';
	be += 2;
	
    if(new != 7)
    {
    	static char Swap[] = "04261537";
    	
    	int pp=0;
    	
    	if((old&0x80) > (new&0x80)
    	|| (old&0x08) > (new&0x08))
    	{
    		*be++ = '0'; pp=1;
    		old = 7;
    	}    	
    	
    	
    	if((new&0x08) && !(old&0x08)){if(pp)*be++ = ';'; *be++ = '1';pp=1;}
    	if((new&0x80) && !(old&0x80)){if(pp)*be++ = ';'; *be++ = '5';pp=1;}
       	
       	if((new&0x70) != (old&0x70))
       	{
   	   		if(pp)*be++ = ';';
   	   		*be++ = '4';
   	   		*be++ = Swap[(new>>4)&7];
       		pp=1;
       	}
       	
      	if((new&7) != (old&7))
      	{
       		if(pp)*be++ = ';';
       		*be++ = '3';
       		*be++ = Swap[new&7];
       	}
       	*be = 0;
    }
    
    be[0] = 'm';
    be[1] = 0;
    
    return Buff;
}
#endif

static void FlushSetAttr()
{
    #ifdef DJGPP

	if(Colors)textattr(TextAttr);

    #else
    
	Buffaa(ColorDiff(OldAttr, TextAttr));

    #endif
    
	OldAttr = TextAttr;
}

static void UpDown(char *s, int oldy,int newy)
{
	/***/if(newy == oldy-1)sprintf(s, "\33[A");
	else if(newy == oldy+1)sprintf(s, "\33[B");
	else if(newy < oldy)sprintf(s, "\33[%dA", oldy-newy);
	else if(newy > oldy)sprintf(s, "\33[%dB", newy-oldy);
	else *s=0;
}

static void LeftRight(char *s, int oldx,int newx)
{
	/***/if(newx == oldx+1)sprintf(s, "\33[C");
	else if(newx == oldx-1)sprintf(s, "\33[D");
	else if(newx < oldx)sprintf(s, "\33[%dD", oldx-newx);
	else if(newx > oldx)sprintf(s, "\33[%dC", newx-oldx);
	else *s=0;
}

/* Permanent field - required for correctly removing the timered items */
static int Field[24][40];
static int Color[24][40];

/* Currently visible field - required for optimized terminal output */
static int VField[24][40];
static int VColor[24][40];

/* Requested visible field - readable by other modules too */
int QField[24][40];
int QColor[24][40];

static char *LocationDiff(int oldx,int oldy, int newx,int newy, int color)
{
	static char Buff[24];
	char *s=Buff;
	
	register int xa = XA;
	register int ya = YA;
	
	oldx += xa;
	oldy += ya;
	
	newx += xa;
	newy += ya;
	
	Buff[0]=0;
	
	if(newx==oldx && newy==oldy)return Buff;
	
	if(!newx)
	{
		if(newy==oldy+4) return strcpy(Buff, "\n\n\n\n");
		if(newy==oldy+3) return strcpy(Buff, "\n\n\n");
		if(newy==oldy+2) return strcpy(Buff, "\n\n");
		if(newy==oldy+1) return strcpy(Buff, "\n");
		UpDown(s, oldy,newy);
		return strcat(Buff, "\r");
	}
	
	#if 0
	if(newx==oldx-2)
	{
		*s++ = '\b';
		goto Da;
	}
	#endif
	if(newx==oldx-1)
	{
	#if 0
	Da:	
	#endif
		*s++ = '\b';
		UpDown(s, oldy,newy);
		return Buff;
	}
	if(newy == oldy)
	{
		if(newx>oldx && newx<oldx+3)
		{
			int a;
			for(a=oldx; a<newx; a++)
				if(QColor[newy-ya][a-xa] != color)
					break;
			if(a==newx) /* yeah! */
			{
				for(a=oldx; a<newx; a++)
					*s++ = QField[newy-ya][a-xa];
				*s=0;
				return Buff;
			}
		}
	
		LeftRight(s, oldx,newx);
		return Buff;
	}
	sprintf(Buff, "\33[%d;%df", newy+1, newx+1);
	return Buff;
}

static void Locate(int x, int y)
{
	Buffaa(LocationDiff(WhereX,WhereY, x,y, TextAttr));
	WhereX = x;
	WhereY = y;
}

static void PutChar(int ch)
{
	*OutputTail++ = ch;
	
	WhereX++;
}

static int ch2char(int c)
{
	static char Mrks[16] = "#\"-'-`-^.|,<.>v+";
	                     /* 0 123456789ABCDEF          */
	                     /* &1=up, &2=lt, &4=rt, &8=dn */
		
	if(!iscorner(c))
	{
		/* We don't print confusing chars. */
		if(c < ' ')c = ' ';
		return c;
	}
	
	return Mrks[uncorner(c)];
}

struct Job
{
	int Done;
	int x, y, ch, color;
};

static void FlushScreen(void)
{
	int x, y, JobCount=0, JobPos, JobsLeft;
	
	struct Job *Jobs;
	
	/* Is there a problem? */
	#define Problem(x,y) \
			(QField[y][x] != VField[y][x]  \
		  || QColor[y][x] != VColor[y][x])
		  
	/* There is problem, if either color or character is different */
	
	for(y=0; y<24; y++)
		for(x=0; x<40; x++)
			if(Problem(x,y))
				JobCount++;
			
	if(!JobCount)
	{
		/* Let's have a holiday. */
		return;
	}
		
	/* For entire screen redraw, Jobs can take              *
	 * sizeof(struct Job) * 40 * 24 bytes of memory.        *
	 * That is, 40*24 * 5*4, which is 40*24*20, about 19kB. */
	
	JobPos = 0;
	Jobs = (struct Job *)malloc(JobCount * (sizeof(struct Job)));
	if(!Jobs)
	{
		DoneVideo();
		printf("Out of memory in FlushScreen in "__FILE__);
		exit(ENOMEM);
	}
	
	for(y=0; y<24; y++)
		for(x=0; x<40; x++)
		{
			/* Is there a problem? */
			if(Problem(x,y))
			{
				/* Yes, we need to fix it,   *
				 * but let's handle it later */
				Jobs[JobPos].x = x;
				Jobs[JobPos].y = y;
				Jobs[JobPos].Done = 0;
				Jobs[JobPos].ch    = QField[y][x];
				Jobs[JobPos].color = QColor[y][x];
				JobPos++;
			}
		}
		
	/* Ok, now the most exciting part. */
	
	for(JobsLeft=JobCount; JobsLeft; JobsLeft--)
	{
		int x,y, a, bp=0;
		char score=100;
		
		/* We must select now the job which *
		 * bestly fits in our motivation... */
		 
		if(JobCount > 40*4) /* We should not use CPU much, however */
		{
			for(a=0; a<JobCount; a++)
				if(!Jobs[a].Done)
				{
					bp = a;
					break;
				}
		}
		else
			for(a=0; a<JobCount; a++)
				if(!Jobs[a].Done)
				{
					char Score = ColorDiffLen[TextAttr&255][Jobs[a].color&255];
					
					if(VField[Jobs[a].y][Jobs[a].x] == Jobs[a].ch
					&& VColor[Jobs[a].y][Jobs[a].x] == Jobs[a].color)
					{
						Jobs[a].Done=1;
						goto Haa;
					}
					
					Score += Jobs[a].x > Jobs[bp].x;
					Score += Jobs[a].y > Jobs[bp].y;
					
					if(Score >= score)continue;
					
					Score += (char)strlen(
						LocationDiff(
							WhereX,WhereY,
							Jobs[a].x,Jobs[a].y,
							Jobs[a].color));
					
					if(Score < score)
					{
						score = Score;
						bp    = a;
						if(!score)break;	/* The best was found */
					}
				}
				
		x = Jobs[bp].x;
		y = Jobs[bp].y;
			
		/* Then execute it. */
		SetAttr(VColor[y][x] = Jobs[bp].color);
		FlushSetAttr();
		Locate(x, y);
		PutChar(VField[y][x] = Jobs[bp].ch);
		
		/* Mark it done. */
		Jobs[bp].Done = 1;
		
		#if EVERYFLUSH
		
		Flush();
		
		#endif
	Haa:;
	}
	
	/* They don't need to be prisoned... */	
	free(Jobs);
}

/* A temporary object */
static struct Temp
{
	int x, y, ch, color, time;
	struct Temp *next;
} *Temps = NULL;

static void AddTemporary(int x, int y, int ch, int color, int time)
{
	struct Temp *tmp;
	
	tmp = (struct Temp *)malloc(sizeof(struct Temp));
	if(!tmp)
	{
		DoneVideo();
		printf("Out of memory in AddTemporary in "__FILE__);
		exit(ENOMEM);
	}
	
	tmp->x = x;
	tmp->y = y;
	tmp->ch = ch;
	tmp->time = time;
	tmp->color = color;
	tmp->next = Temps;
	
	Temps = tmp;
}

void ClrScr(void)
{
	int x, y;
	
	for(y=0; y<24; y++)
		for(x=0; x<40; x++)
		{
			struct Temp *tmp;
			
			Field[y][x] = ' ';
			Color[y][x] = 7;
			
			for(tmp=Temps; tmp; tmp=tmp->next)
				if(tmp->x==x && tmp->y==y)
				{
					goto Noway;
					/* May not clear it from screen */
				}

			QField[y][x] = ' ';
			QColor[y][x] = 7;
			
		Noway:;
		}
}

void RefreshScreen(int x)
{
	int y;
	
	GetWinSize();
	
	OldAttr = -1;
	
	Flush();
	
	SetAttr(7);
	FlushSetAttr();
	Buffaa("\33[2J\33[H"); /* Clear screen, home cursor */
	
	#if HIDE_CURSOR	
	Buffaa("\33[?25l");		/* Hide cursor */
	#endif
	
	Flush();
	
	WhereX = -XA;
	WhereY = -YA;
	
	for(y=0; y<24; y++)
		for(x=0; x<40; x++)
		{
			VField[y][x] = ' ';
			VColor[y][x] = TextAttr;
		}
}

void InitVideo(void)
{
	int x, y;
	
	InitCons();
	
	if(LINES<24 || COLS<40)
	{
		printf(
			"Sorry, your terminal geometry is not big enough.\n"
			"40x24 required at least.\n");
		exit(0);
	}
	
	for(x=0; x<256; x++)
		for(y=0; y<256; y++)
			ColorDiffLen[x][y] = (char)strlen(ColorDiff(x, y));
	
	ClrScr();
	
	#ifdef SIGWINCH
	 signal(SIGWINCH, RefreshScreen);
	#endif
		
	RefreshScreen(0);
}

/* transp: 0=not transparent - 1=transparent - 2=over nonpermanent also
 *   time: 0=permanent       - other=disappears after this count of ticks
 *     ch: ascii character or corner combination
 *  color: combination of 1=blue, 2=green, 4=red, 8=intense
 */
void SetChar(int x, int y, int ch, int color, int transp, int time)
{
	/* Check against bounds */
	if(x<0 || x>39 || y<0 || y>23)return;
	
	/* For vt100, we don't care about the transparency flag,
	 *            because the terminal simply does not support it...
	 */
	transp = transp;
	
	ch = ch2char(ch);
	
	if(!time)
	{
		struct Temp *tmp;
		
		/* immediately */
		Field[y][x] = ch;
		Color[y][x] = color;
		
		if(transp != 2)
		{
			for(tmp=Temps; tmp; tmp=tmp->next)
				if(tmp->x==x && tmp->y==y)return;
			/* May not put it on it */
		}
		else
		{
			struct Temp **Owner = &Temps;
			for(;;)
			{
				if(!(tmp = *Owner))break;
				if(tmp->x==x && tmp->y==y)
				{
					/* kill it! */
					struct Temp *tmp2 = tmp->next;
					free(tmp);
					*Owner = tmp2;
				}
				else
					Owner = &tmp->next;
			}
		}
	}
	else
	{
		/* This shouldn't ever happen,
		 * but let's doublesecure. */
		if(time < 0)return;
		
		/* Temporarily */
		AddTemporary(x, y, ch, color, time);
	}
	
	QField[y][x] = ch;
	QColor[y][x] = color;
}

static void Evaluate(void)
{
	struct Temp **Owner = &Temps;
	
	for(;;)
	{
		struct Temp *tmp = *Owner;
		if(!tmp)break;
		
		/* Is it at the end of its lifetime? */
		if(!--(tmp->time))
		{
			int x=tmp->x, y=tmp->y;
			
			/* Delete it from the list */
			struct Temp *tmp2 = tmp->next;
			free(tmp);
			*Owner = tmp2;

			/* Find the correct underground for this */			
			for(tmp2=Temps; tmp2; tmp2=tmp2->next)
				if(x==tmp2->x && y==tmp2->y)
				{
					QField[y][x] = tmp2->ch;
					QColor[y][x] = tmp2->color;
					break;
				}
				
			/* If not found, take it from Field.      *
			 * This is the most often case, actually. */
			if(!tmp2)
			{
				QField[y][x] = Field[y][x];
				QColor[y][x] = Color[y][x];
			}
			
			continue;
		}
		
		Owner = &tmp->next;
	}
}

void UpdateVideo(int numticks)
{
	while(numticks)
	{
		FlushScreen();
	
		/* After the screen has been draw,
		 * prepare already for the next frame..
		 */
		Evaluate();
		 
		numticks--;
	}
	
	Flush();
}

void DoneVideo(void)
{
	SetAttr(7);
	FlushSetAttr();

#if CLEAR_SCREEN_AT_END
	Buffaa("\33[2J\33[H"); /* Clear screen */
#else
	Locate(0, 24);
	Buffaa("\n");
#endif

#if HIDE_CURSOR
	Buffaa("\33[?25h");
#endif

	Flush();
	
	DoneCons();
}
